from pysolarfocus.components.base.component import Component
from pysolarfocus.components.base.data_value import DataValue


def _assign_absolute_addresses(data_values: list[DataValue], start_address):
    for data_value in data_values:
        data_value.absolut_address = start_address


def test_calculate_ranges_handles_simple_case():
    absolute_address = 500
    data_values = [
        DataValue(address=0),
        DataValue(address=1),
        DataValue(address=2),
        DataValue(address=3),
        DataValue(address=4),
    ]
    _assign_absolute_addresses(data_values, absolute_address)
    slices = Component._calculate_ranges(data_values)
    assert len(slices) == 1
    assert slices[0].absolute_address == absolute_address
    assert slices[0].relative_address == 0
    assert slices[0].count == 5


def test_calculate_ranges_handles_skipped_values():
    # Tests if 503 is skipped
    absolute_address = 500
    data_values = [
        DataValue(address=0),
        DataValue(address=1),
        DataValue(address=2),
        DataValue(address=4),
        DataValue(address=5),
    ]
    _assign_absolute_addresses(data_values, absolute_address)
    slices = Component._calculate_ranges(data_values)
    assert len(slices) == 2
    assert slices[0].absolute_address == absolute_address
    assert slices[0].relative_address == 0
    assert slices[0].count == 3
    assert slices[1].absolute_address == 504
    assert slices[1].relative_address == 4
    assert slices[1].count == 2


def test_calculate_ranges_handles_multiregister_values():
    absolute_address = 500
    data_values = [
        DataValue(address=0, count=2),
        DataValue(address=2, count=3),
        DataValue(address=5),
    ]
    _assign_absolute_addresses(data_values, absolute_address)
    slices = Component._calculate_ranges(data_values)
    assert len(slices) == 1
    assert slices[0].absolute_address == absolute_address
    assert slices[0].relative_address == 0
    assert slices[0].count == 6


def test_calculate_ranges_handles_multiregister_and_skipped_values():
    absolute_address = 500
    data_values = [
        DataValue(address=0, count=2),
        DataValue(address=2, count=3),
        DataValue(address=20, count=10),
    ]
    _assign_absolute_addresses(data_values, absolute_address)
    slices = Component._calculate_ranges(data_values)
    assert len(slices) == 2
    assert slices[0].absolute_address == absolute_address
    assert slices[0].relative_address == 0
    assert slices[0].count == 5
    assert slices[1].absolute_address == absolute_address + 20
    assert slices[1].relative_address == 20
    assert slices[1].count == 10


def test_calculate_ranges_handles_multiple_skipped_values():
    # Test the Heating Circuit configuration
    absolute_address = 32600
    data_values = [
        DataValue(address=0),
        DataValue(address=2),
        DataValue(address=3),
        DataValue(address=5),
        DataValue(address=6),
        DataValue(address=7),
    ]
    _assign_absolute_addresses(data_values, absolute_address)
    slices = Component._calculate_ranges(data_values)
    assert len(slices) == 3

    assert slices[0].absolute_address == absolute_address
    assert slices[0].relative_address == 0
    assert slices[0].count == 1

    assert slices[1].absolute_address == absolute_address + 2
    assert slices[1].relative_address == 2
    assert slices[1].count == 2

    assert slices[2].absolute_address == absolute_address + 5
    assert slices[2].relative_address == 5
    assert slices[2].count == 3



"""Solarfocus pelletsboiler component"""

from .. import ApiVersions, Systems
from .base.component import Component
from .base.data_value import DataValue
from .base.enums import DataTypes, RegisterTypes


class BiomassBoiler(Component):
    def __init__(self, input_address=2400, holding_address=-1, api_version: ApiVersions = ApiVersions.V_23_010, system=Systems.ECOTOP) -> None:

        if api_version.greater_or_equal(ApiVersions.V_22_090.value):
            holding_address = 33400

        super().__init__(input_address, holding_address)
        self.temperature = DataValue(address=0, multiplier=0.1)
        self.status = DataValue(address=1, data_type=DataTypes.UINT)
        self.time_of_operation_at_maintenance = DataValue(address=2, count=2, data_type=DataTypes.INT)

        self.message_number = DataValue(address=4)
        self.door_contact = DataValue(address=5)
        self.cleaning = DataValue(address=6)
        self.ash_container = DataValue(address=7)
        self.outdoor_temperature = DataValue(address=8, multiplier=0.1)
        self.boiler_operating_mode = DataValue(address=9)
        self.octoplus_buffer_temperature_bottom = DataValue(address=10, multiplier=0.1)
        self.octoplus_buffer_temperature_top = DataValue(address=11, multiplier=0.1)

        if api_version.greater_or_equal(ApiVersions.V_22_090.value) and system is not Systems.ECOTOP:
            self.log_wood = DataValue(address=12, data_type=DataTypes.UINT)
            self.sweep_function_start_stop = DataValue(address=10, register_type=RegisterTypes.HOLDING)
            self.sweep_function_extend = DataValue(address=11, register_type=RegisterTypes.HOLDING)

        if api_version.greater_or_equal(ApiVersions.V_23_010.value):
            self.pellet_usage_last_fill = DataValue(address=14, count=2, multiplier=0.1)
            self.pellet_usage_total = DataValue(address=16, count=2, multiplier=0.1)
            self.heat_energy_total = DataValue(address=18, count=2, multiplier=0.1)

            self.pellet_usage_reset = DataValue(address=12, register_type=RegisterTypes.HOLDING)

        if api_version.greater_or_equal(ApiVersions.V_23_080.value):
            self.sweep_almost_done = DataValue(address=20)



"""Solarfocus boiler component"""
from .. import ApiVersions
from .base.component import Component
from .base.data_value import DataValue
from .base.enums import DataTypes, RegisterTypes


class Boiler(Component):
    def __init__(self, input_address: int = 500, holding_address: int = 32000, api_version: ApiVersions = ApiVersions.V_21_140) -> None:
        super().__init__(input_address=input_address, holding_address=holding_address)
        self.temperature = DataValue(address=0, multiplier=0.1)
        self.state = DataValue(address=1, data_type=DataTypes.UINT)
        self.mode = DataValue(address=2, data_type=DataTypes.UINT)

        self.target_temperature = DataValue(address=0, multiplier=10, register_type=RegisterTypes.HOLDING)
        self.single_charge = DataValue(address=1, register_type=RegisterTypes.HOLDING)
        self.holding_mode = DataValue(address=2, register_type=RegisterTypes.HOLDING)

        if api_version.greater_or_equal(ApiVersions.V_20_110.value):
            self.circulation = DataValue(address=3, register_type=RegisterTypes.HOLDING)



"""Solarfocus performance calculator"""

from .part import Part


class PerformanceCalculator(Part):
    """
    Performing performance calculations
    """

    nominator: Part
    denominator: Part
    value: int = None

    def __init__(self, nominator: Part, denominator: Part) -> None:
        self.nominator = nominator
        self.denominator = denominator

    @property
    def value(self) -> float:
        """
        Return value of the calculation
        """
        if self.denominator.scaled_value:
            return self.nominator.scaled_value / self.denominator.scaled_value
        return 0.0

    @property
    def scaled_value(self) -> float:
        """
        To provide a uniform interface for integrating, a scaled value property is offered
        """
        return self.value



"""Solarfocus photovoltaic component"""
from .. import ApiVersions
from .base.component import Component
from .base.data_value import DataValue
from .base.enums import RegisterTypes


class Photovoltaic(Component):
    def __init__(self, api_version: ApiVersions = ApiVersions.V_21_140) -> None:
        super().__init__(input_address=2500, holding_address=33407)
        self.power = DataValue(address=0, count=2)
        self.house_consumption = DataValue(address=2, count=2)
        self.heatpump_consumption = DataValue(address=4, count=2)
        self.grid_import = DataValue(address=6, count=2)
        self.grid_export = DataValue(address=8, count=2)

        if api_version.greater_or_equal(ApiVersions.V_21_140.value):
            self.overcharge_possible = DataValue(address=10, count=1)
            self.overcharge_active = DataValue(address=11, count=1)

        self.smart_meter = DataValue(address=0, register_type=RegisterTypes.HOLDING)
        self.photovoltaic = DataValue(address=1, register_type=RegisterTypes.HOLDING)
        self.grid_im_export = DataValue(address=2, register_type=RegisterTypes.HOLDING)



"""Solarfocus heat pump component"""

from .base.component import Component
from .base.data_value import DataValue
from .base.enums import DataTypes, RegisterTypes
from .base.performance_calculator import PerformanceCalculator


class HeatPump(Component):
    def __init__(self) -> None:
        super().__init__(input_address=2300, holding_address=33404)
        self.supply_temperature = DataValue(address=0, multiplier=0.1)
        self.return_temperature = DataValue(address=1, multiplier=0.1)
        self.flow_rate = DataValue(address=2)
        self.compressor_speed = DataValue(address=3)
        self.evu_lock_active = DataValue(address=4, data_type=DataTypes.UINT)
        self.defrost_active = DataValue(address=5, data_type=DataTypes.UINT)
        self.boiler_charge = DataValue(address=6, data_type=DataTypes.UINT)
        self.thermal_energy_total = DataValue(address=7, count=2, multiplier=0.001)
        self.thermal_energy_drinking_water = DataValue(address=9, count=2, multiplier=0.001)
        self.thermal_energy_heating = DataValue(address=11, count=2, multiplier=0.001)
        self.electrical_energy_total = DataValue(address=13, count=2, multiplier=0.001)
        self.electrical_energy_drinking_water = DataValue(address=15, count=2, multiplier=0.001)
        self.electrical_energy_heating = DataValue(address=17, count=2, multiplier=0.001)
        self.electrical_power = DataValue(address=19)
        self.thermal_power_cooling = DataValue(address=20)
        self.thermal_power_heating = DataValue(address=21)
        self.thermal_energy_cooling = DataValue(address=22, count=2, multiplier=0.001, data_type=DataTypes.UINT)
        self.electrical_energy_cooling = DataValue(address=24, count=2, multiplier=0.001, data_type=DataTypes.UINT)

        self.vampair_state = DataValue(address=26, data_type=DataTypes.UINT)

        # This register is shared with the biomass boiler
        self.outdoor_temperature = DataValue(address=108, multiplier=0.1)

        self.evu_lock = DataValue(address=0, register_type=RegisterTypes.HOLDING)
        self.smart_grid = DataValue(address=1, register_type=RegisterTypes.HOLDING)
        self.outdoor_temperature_external = DataValue(address=2, multiplier=10, register_type=RegisterTypes.HOLDING)

        self.cop_heating = PerformanceCalculator(self.thermal_power_heating, self.electrical_power)
        self.cop_cooling = PerformanceCalculator(self.thermal_power_cooling, self.electrical_power)
        self.performance_overall = PerformanceCalculator(self.thermal_energy_total, self.electrical_energy_total)
        self.performance_overall_heating = PerformanceCalculator(self.thermal_energy_heating, self.electrical_energy_heating)
        self.performance_overall_drinking_water = PerformanceCalculator(self.thermal_energy_drinking_water, self.electrical_energy_drinking_water)



"""Solarfocus component factory"""
from . import ApiVersions, Systems
from .components.biomass_boiler import BiomassBoiler
from .components.boiler import Boiler
from .components.buffer import Buffer, TherminatorBuffer
from .components.fresh_water_module import FreshWaterModule
from .components.heat_pump import HeatPump
from .components.heating_circuit import HeatingCircuit, TherminatorHeatingCircuit
from .components.photovoltaic import Photovoltaic
from .components.solar import Solar
from .modbus_wrapper import ModbusConnector


class ComponentFactory:
    def __init__(self, modbus_connector: ModbusConnector) -> None:
        self.__modbus_connector = modbus_connector

    def heating_circuit(self, system: Systems, count: int, api_version: ApiVersions) -> list[HeatingCircuit]:
        input_addresses = list(range(1100, 1100 + (50 * count), 50))
        holding_addresses = list(range(32600, 32600 + (50 * count), 50))
        heating_circuits = []
        for i in range(count):
            input, holding = input_addresses[i], holding_addresses[i]
            if system in [Systems.THERMINATOR, Systems.ECOTOP]:
                heating_circuit = TherminatorHeatingCircuit(input, holding, api_version=api_version).initialize(self.__modbus_connector)
            else:
                heating_circuit = HeatingCircuit(input, holding, api_version=api_version).initialize(self.__modbus_connector)
            heating_circuits.append(heating_circuit)
        return heating_circuits

    def boiler(self, system: Systems, count: int, api_version: ApiVersions) -> list[Boiler]:
        input_addresses = list(range(500, 500 + (50 * count), 50))
        holding_addresses = list(range(32000, 32000 + (50 * count), 50))
        boilers = []
        for i in range(count):
            input, holding = input_addresses[i], holding_addresses[i]
            boilers.append(Boiler(input, holding, api_version=api_version).initialize(self.__modbus_connector))
        return boilers

    def buffer(self, system: Systems, count: int, api_version: ApiVersions) -> list[Buffer]:
        input_addresses = list(range(1900, 1900 + (20 * count), 20))
        holding_addresses = -1
        if api_version.greater_or_equal(ApiVersions.V_22_090.value):
            holding_addresses = list(range(34000, 34000 + (50 * count), 50))
        buffers = []
        for i in range(count):
            input = input_addresses[i]
            if api_version.greater_or_equal(ApiVersions.V_22_090.value):
                holding = holding_addresses[i]
            else:
                holding = -1
            if system in [Systems.THERMINATOR, Systems.ECOTOP]:
                buffer = TherminatorBuffer(input).initialize(self.__modbus_connector)
            else:
                buffer = Buffer(input, holding, api_version=api_version).initialize(self.__modbus_connector)
            buffers.append(buffer)
        return buffers

    def fresh_water_modules(self, system: Systems, count: int, api_version: ApiVersions) -> list[FreshWaterModule]:
        input_addresses = list(range(700, 700 + (25 * count), 25))
        fresh_water_modules = []
        for i in range(count):
            input = input_addresses[i]
            fresh_water_modules.append(FreshWaterModule(input, api_version=api_version).initialize(self.__modbus_connector))
        return fresh_water_modules

    def heatpump(self, system: Systems) -> HeatPump:
        return HeatPump().initialize(self.__modbus_connector)

    def photovoltaic(self, system: Systems, api_version: ApiVersions) -> Photovoltaic:
        return Photovoltaic(api_version=api_version).initialize(self.__modbus_connector)

    def pelletsboiler(self, system: Systems, api_version: ApiVersions) -> BiomassBoiler:
        return BiomassBoiler(api_version=api_version, system=system).initialize(self.__modbus_connector)

    def solar(self, system: Systems) -> Solar:
        return Solar().initialize(self.__modbus_connector)



"""Solarfocus fresh water module component"""
from .. import ApiVersions
from .base.component import Component
from .base.data_value import DataValue


class FreshWaterModule(Component):
    def __init__(self, input_address=700, api_version: ApiVersions = ApiVersions.V_23_020) -> None:
        super().__init__(input_address)
        self.state = DataValue(address=0)

        if api_version.greater_or_equal(ApiVersions.V_23_040.value):
            self.supply_temperature = DataValue(address=1, count=1, multiplier=0.1)
            self.flow_rate = DataValue(address=2, count=1)
            self.target_temperature = DataValue(address=3, count=1, multiplier=0.1)
            self.valve = DataValue(address=4, count=1)



"""Solarfocus buffer component"""

from .. import ApiVersions
from .base.component import Component
from .base.data_value import DataValue
from .base.enums import DataTypes, RegisterTypes


class Buffer(Component):
    def __init__(self, input_address: int = 1900, holding_address=-1, api_version: ApiVersions = ApiVersions.V_21_140) -> None:

        super().__init__(input_address, holding_address)

        self.top_temperature = DataValue(address=0, multiplier=0.1)
        self.bottom_temperature = DataValue(address=1, multiplier=0.1)
        self.pump = DataValue(address=2)
        self.state = DataValue(address=3, data_type=DataTypes.UINT)
        self.mode = DataValue(address=4, data_type=DataTypes.UINT)

        if api_version.greater_or_equal(ApiVersions.V_22_090.value):
            self.external_top_temperature_x44 = DataValue(address=0, multiplier=10, register_type=RegisterTypes.HOLDING)
            self.external_middle_temperature_x36 = DataValue(address=1, multiplier=10, register_type=RegisterTypes.HOLDING)
            self.external_bottom_temperature_x35 = DataValue(address=2, multiplier=10, register_type=RegisterTypes.HOLDING)


class TherminatorBuffer(Buffer):
    def __init__(self, address: int = 1900) -> None:
        super().__init__(input_address=address)
        self.x35_temperature = DataValue(address=2, multiplier=0.1)
        self.pump = DataValue(address=3)
        self.state = DataValue(address=4, data_type=DataTypes.UINT)
        self.mode = DataValue(address=5, data_type=DataTypes.UINT)



"""Solarfocus abstract component"""
import logging

from ...modbus_wrapper import ModbusConnector
from .data_value import DataValue
from .enums import DataTypes, RegisterTypes
from .performance_calculator import PerformanceCalculator
from .register_slice import RegisterSlice


class Component:
    def __init__(self, input_address: int, holding_address: int = -1) -> None:
        self.input_address = input_address
        self.input_count = 0
        self.holding_address = holding_address
        self.holding_count = 0
        self.__data_values = None
        self.__input_values = None
        self.__holding_values = None
        self.__performance_calculators = None
        self.__modbus = None

    def initialize(self, modbus: ModbusConnector):
        """
        Initializes the absolute addresses of the DataValues and the count of the Component
        """

        self.__modbus = modbus

        for _, value in self.__get_data_values():

            if value.register_type == RegisterTypes.INPUT:
                value.absolut_address = self.input_address
            else:
                value.absolut_address = self.holding_address
                # Holding registers can write to the heating system
                value.modbus = modbus

        # Dynamically calcuate how many registers have to be read
        if len(self.__get_input_values()) > 0:
            _, last_input_value = self.__get_input_values()[-1]
            self.input_count = last_input_value.address + last_input_value.count

        if len(self.__get_holding_values()) > 0:
            _, last_holding_value = self.__get_holding_values()[-1]
            self.holding_count = last_holding_value.address + last_holding_value.count

        # Calculate the address slices we need to read
        # This is necessary because the modbus protocol can block the read/write of some registers
        # => if one of these registers is between the start and end address of the read/write we need to skip it
        self.__input_slices = Component._calculate_ranges([v for (n, v) in self.__get_input_values()])
        self.__holding_slices = Component._calculate_ranges([v for (n, v) in self.__get_holding_values()])

        # Initialize has_performance_calculators
        self.__get_performance_calculators()

        return self

    @property
    def holding_slices(self) -> list[RegisterSlice]:
        """
        Returns the address slices of the holding registers
        """
        return self.__holding_slices

    @property
    def input_slices(self) -> list[RegisterSlice]:
        """
        Returns the address slices of the input registers
        """
        return self.__input_slices

    @staticmethod
    def _calculate_ranges(datavalues: list[DataValue]) -> list[RegisterSlice]:
        slices = []
        if len(datavalues) < 1:
            return slices
        current_slice = RegisterSlice(datavalues[0].get_absolute_address(), datavalues[0].address, datavalues[-1].address + datavalues[-1].count)
        for i, datavalue in enumerate(datavalues[:-1]):
            if datavalue.address + datavalue.count != datavalues[i + 1].address:
                # A gap was found => define the range from the start of the current slice to the end of the current address+count
                current_slice.count = (datavalue.address + datavalue.count) - current_slice.relative_address
                slices.append(current_slice)
                # the new count is the distance between the next address and the last address
                new_count = (datavalues[-1].address + datavalues[-1].count) - datavalues[i + 1].address
                current_slice = RegisterSlice(datavalues[i + 1].get_absolute_address(), datavalues[i + 1].address, new_count)
        slices.append(current_slice)
        return slices

    @property
    def has_input_address(self) -> bool:
        return self.input_address >= 0 and self.input_count > 0

    @property
    def has_holding_address(self) -> bool:
        return self.holding_address >= 0 and self.holding_count > 0

    @property
    def has_performance_calculators(self) -> bool:
        return len(self.__get_performance_calculators()) > 0

    def __get_data_values(self) -> list[tuple[str, DataValue]]:
        """
        Returns all DataValues of this Component
        """
        if self.__data_values is None:
            self.__data_values = [(k, v) for k, v in self.__dict__.items() if isinstance(v, DataValue)]
        return self.__data_values

    def __get_input_values(self) -> list[tuple[str, DataValue]]:
        """
        Get sorted list of all DataValues with RegisterType Input
        """
        if self.__input_values is None:
            self.__input_values = [(k, v) for k, v in self.__get_data_values() if v.register_type == RegisterTypes.INPUT]
            self.__input_values = sorted(self.__input_values, key=lambda item: item[1].address)
        return self.__input_values

    def __get_holding_values(self) -> list[tuple[str, DataValue]]:
        """
        Get sorted list of all DataValues with RegisterType Holding
        """
        if self.__holding_values is None:
            self.__holding_values = [(k, v) for k, v in self.__get_data_values() if v.register_type == RegisterTypes.HOLDING]
            self.__holding_values = sorted(self.__holding_values, key=lambda item: item[1].address)
        return self.__holding_values

    def __get_performance_calculators(self) -> list[tuple[str, PerformanceCalculator]]:
        """
        Returns all PerformanceCalculators of this Component
        """
        if self.__performance_calculators is None:
            self.__performance_calculators = [(k, v) for k, v in self.__dict__.items() if isinstance(v, PerformanceCalculator)]
        return self.__performance_calculators

    @staticmethod
    def __unsigned_to_signed(number: int, byte_count: int) -> int:
        return int.from_bytes(number.to_bytes(byte_count, "little", signed=False), "little", signed=True)

    def update(self) -> bool:
        """
        Retrieve current values from the heating system
        """
        failed = False
        if self.has_input_address:
            read_success, registers = self.__modbus.read_input_registers(self.input_slices, self.input_count)
            parsing_success = False
            if read_success:
                parsing_success = self._parse(registers, RegisterTypes.INPUT) and read_success
            failed = not parsing_success and read_success or failed
            if failed:
                logging.error(f"Failed to read input registers of {self.__class__.__name__}")

        if self.has_holding_address:
            read_success, registers = self.__modbus.read_holding_registers(self.holding_slices, self.holding_count)
            parsing_success = False
            if read_success:
                parsing_success = self._parse(registers, RegisterTypes.HOLDING) and read_success
            failed = not (parsing_success and read_success) or failed
            if failed:
                logging.error(f"Failed to read holding registers of {self.__class__.__name__}")
        return not failed

    def _parse(self, data: list[int], type: RegisterTypes) -> bool:
        """
        Dynamically assignes the values to the DataValues of this Component
        """
        if len(data) != (self.input_count if type == RegisterTypes.INPUT else self.holding_count):
            logging.error(
                f"Data length does not match the expected length of {self.input_count if type == RegisterTypes.INPUT else self.holding_count} for {self.__class__.__name__}"
            )
            return False

        encountered_error = False
        for name, value in self.__get_input_values() if type == RegisterTypes.INPUT else self.__get_holding_values():
            try:
                # Multi-register values (UINT32, INT32)
                if value.count == 2:
                    _value = (data[value.address] << 16) + data[value.address + 1]
                else:
                    _value = data[value.address]

                # Datatype
                if value.data_type == DataTypes.INT:
                    _value = Component.__unsigned_to_signed(_value, value.count * 2)

                # Store
                value.value = _value
            except:
                logging.exception(f"Error while parsing {name} of {self.__class__.__name__}")
                encountered_error = True
        return not encountered_error

    def __repr__(self) -> str:
        message = ["=" * 12]
        message.append(f"{self.__class__.__name__}")
        message.append("=" * 12)
        if self.has_input_address:
            message.append("---Input:")
            for name, value in self.__get_input_values():
                message.append(f"{name} | raw:{value.value} scaled:{value.scaled_value}")
        if self.has_holding_address:
            message.append("---Holding:")
            for name, value in self.__get_holding_values():
                message.append(f"{name} | raw:{value.value} scaled:{value.scaled_value}")
        if self.has_performance_calculators:
            message.append("---Calculations:")
            for name, value in self.__get_performance_calculators():
                message.append(f"{name} | raw:{value.value} scaled:{value.scaled_value}")
        return "\n".join(message)



from pysolarfocus import ApiVersions, SolarfocusAPI, Systems

# Create the Solarfocus API client
solarfocus = SolarfocusAPI(
    ip="solarfocus", system=Systems.VAMPAIR, api_version=ApiVersions.V_23_020  # adapt IP-Address  # change to Systems.Therminator
)  # select Solarfocus version

if not solarfocus.connect():
    print("Connecting to solarfocus failed.")
    exit(1)

# Fetch the values
if not solarfocus.update():
    print("Updating solarfocus failed.")
    exit(1)

# Print the values
print(solarfocus)
print(solarfocus.heating_circuits[0])
print("\n")
print(solarfocus.boilers[0])
print("\n")
print(solarfocus.buffers[0])
print("\n")
if solarfocus.system in  [Systems.THERMINATOR, Systems.ECOTOP]:
    print(solarfocus.biomassboiler)
if solarfocus.system is Systems.VAMPAIR:
    print(solarfocus.heatpump)
print("\n")
print(solarfocus.photovoltaic)
print("\n")
print(solarfocus.fresh_water_modules[0])
print("\n")
print(solarfocus.solar)
print("\n")
